//
//	Encoder for G.729A CS-ACELP Speech Codec
// 
//		fixed_codebook_search.cpp : Module for 
//			- [A.3.8] Fixed Codebook
//

#include "StdAfx.h"
#include "fixed_codebook_search.h"

/*****************************************************************************
/*
/*	  Function for the fixed codebook search for a subframe
/*    Input: x[40]: target signal
/*			 h[40]: impulse response
/*			 y[40]: filtered adaptive codebook vector
/*			 g_p: adaptive codebook gain
/*           T: Optimal integer part of pitch delay 
/*			 subframe1: flag to specify whether this is the 1st subframe or 2nd subframe
/*						1 means first subframe, 0 means 2nd subframe
/*    Output: *S: Signs of fixed codebook pulses
/*			  *C: Fixed codebook
/*			  c[40]: fixed codebook vector
/*
/****************************************************************************/
void fixed_codebook_search_subframe( float x[40], float h[40], float y[40], float g_p, float beta, int T,
									 int *S, int *C, float c[40])
{
	float x_p[40], d[40];	
	float phi[40][40], phi_p[40][40];
	float CC,C0,C1,C2,E,E0,E1,E2;
	float C2perE, maxC2perE;	
	int i,j,n;
	int m0,m1,m2,m3;
	int optm0,optm1,optm2,optm3;
	bool jx,s0,s1,s2,s3;

	// Equation (49)
	if (T<40)
		for (n=T; n<40; n++)
			h[n] += beta * h[n-T];

	// Equation (50)
	for (n=0; n<40; n++)
		x_p[n] = x[n] - g_p * y[n];

	// Equation (51)
	for (i=0; i<40; i++)
		for (j=i; j<40; j++)
			for (n=j; n<40; n++)			
				phi[i][j] = h[n-i]*h[n-j];

	for (i=0; i<40; i++)
		for (j=0; j<i; j++)
			phi[i][j] = phi[j][i];

	// Equation (52)
	correlation(x_p, h, d);

	// Equation (56)
	for (i=0; i<40; i++)
		for (j=0; j<40; j++)
			phi_p[i][j] = sign(d[i])*sign(d[j])*phi[i][j];
	
	// Equation (57)
	for (i=0; i<40; i++)
		phi_p[i][i] *= 0.5;
					
	optm0 = m0 = 0;
	optm1 = m1 = 1;
	optm2 = m2 = 2;
	optm3 = m3 = 3;
	CC = fabs(d[m0]) + fabs(d[m1]) + fabs(d[m2]) + fabs(d[m3]);
	E = phi_p[m0][m0] + phi_p[m1][m1] + phi_p[m0][m1] + phi_p[m2][m2] + phi_p[m0][m2] 
	  + phi_p[m1][m2] + phi_p[m3][m3] + phi_p[m0][m3] + phi_p[m1][m3] + phi_p[m2][m3];
	maxC2perE = CC*CC/E;

	for (m0=0; m0<=35; m0+=5)
	{		
		C0 = fabs(d[m0]);
		E0 = phi_p[m0][m0];
		for (m1=1; m1<=36; m1+=5)
		{
			C1 = C0 + fabs(d[m1]);
			E1 = E0 + phi_p[m1][m1] + phi_p[m0][m1];
			for (m2=2; m2<=37; m2+=5)
			{
				C2 = C1 + fabs(d[m2]);
				E2 = E1 + phi_p[m2][m2] + phi_p[m0][m2] + phi_p[m1][m2];
				for (m3=3; m3<=38; m3+=5)
				{
					CC = C2 + fabs(d[m3]);
					E = E2 + phi_p[m3][m3] + phi_p[m0][m3] + phi_p[m1][m3] + phi_p[m2][m3];
					// For equation (59), E should be multiplied by 2, but this operation is saved
					//   because the scale factor is the same for all C^2/E
					C2perE = CC*CC/E;
					if (C2perE > maxC2perE)
					{
						optm0 = m0;
						optm1 = m1;
						optm2 = m2;
						optm3 = m3;	
						maxC2perE = C2perE;
					}
				}

				for (m3=4; m3<=39; m3+=5)
				{
					CC = C2 + fabs(d[m3]);
					E = E2 + phi_p[m3][m3] + phi_p[m0][m3] + phi_p[m1][m3] + phi_p[m2][m3];
					// For equation (59), E should be multiplied by 2, but this operation is saved
					//   because the scale factor is the same for all C^2/E
					C2perE = CC*CC/E;
					if (C2perE > maxC2perE)
					{
						optm0 = m0;
						optm1 = m1;
						optm2 = m2;
						optm3 = m3;
						maxC2perE = C2perE;
					}
				}
			}
		}
	}

	jx = (optm3%5==4);
	s0 = (sign(d[optm0])>0);
	s1 = (sign(d[optm1])>0);
	s2 = (sign(d[optm2])>0);
	s3 = (sign(d[optm3])>0);

	*S = s0 + 2*s1 + 4*s2 + 8*s3;
	*C = optm0/5 + 8*(optm1/5) + 64*(optm2/5) + 512*(2*(optm3/5)+jx);

	for (n=0; n<40; n++)
		c[n] = 0;
	c[optm0] = sign(d[optm0]);
	c[optm1] = sign(d[optm1]);
	c[optm2] = sign(d[optm2]);
	c[optm3] = sign(d[optm3]);

	if (T<40)
		for (n=T; n<40; n++)
			c[n] += beta*c[n-T];

}


/*****************************************************************************
/*
/*	  Function for the fixed codebook gain
/*    Input: x[40]: target signal
/*			 y[40]: filtered adaptive codebook vector
/*			 h[40]: impulse response
/*			 g_p: adaptive codebook gain
/*
/*    Output: z[40]: fixed codebook vector convolved with h[n]
/*			  *g_c: fixed codebook gain
/*
/****************************************************************************/
void fixed_codebook_gain(float x[40], float y[40], float z[40], float c[40], float h[40], float g_p, float *g_c)
{
	float A, B, C;
	float e1=0, e2=0, e3=0, e4=0, e5=0, e6=0;
	int n, i;

	for (n=0; n<40; n++)
		for (i=0, z[n]=0; i<=n; i++)
			z[n] += c[i]*h[n-i];

	for (n=0; n<40; n++)
	{
		e1 += x[n]*x[n];
		e2 += y[n]*y[n];
		e3 += z[n]*z[n];
		e4 += x[n]*y[n];
		e5 += x[n]*z[n];
		e6 += y[n]*z[n];
	}

	A = e3; 
	B = 2*(g_p*e6 - e5);
	C = e1 + g_p*g_p*e2 - 2*g_p*e4;
	
	if (A !=0) 
		*g_c = -B/2/A;
	else if (B != 0)
		*g_c = -C/B;
}