//
//	Encoder for G.729A CS-ACELP Speech Codec
//
//		open_loop.cpp : Modules for 
//			- [A.3.3] Perceptual weighting
//			- [A.3.4] Open-loop pitch analysis
//

#include "StdAfx.h"
#include "open_loop.h"

/*********************************************************************/
/*
/*	  Function to do Perceptual weighting [A.3.3]
/*
/*    Input: s[240], a_h[10], start, head
/*    Output: r[40], s_w[320]
/*
/*********************************************************************/
void perceptual_weighting(float s[240], float a_h[10], int start, int head, float r[40], float s_w[320], float ew[10])
{
	float a_p[11];
	float gamma = 1;
	int pos;
	int i, n;

	start = start + 240;
	head = head + 320;

	//Compute the A'(z) = A^(z/gamma)(1-.7z^-1)
	//a_p[n] = gamma*.75*a_h[n] - 0.7*gamma*a_h[n-1];
	//a_p[0] = a_h[0];
	a_p[0] = 1;
	for (n=1; n<11; n++)
	{
		if (n==1)
			a_p[n] = -0.7;
		else
			a_p[n] = -0.7*gamma*a_h[n-2];
		gamma *= .75;
		a_p[n] += gamma*a_h[n-1];
	}
		
	//Compute the residual signal r(n) for n = 0...39
	// Equation (A.3)
	for (n=0; n<40; n++)
	{
		float sum = 0;
		for (i=1; i<=10; i++)
		{
			pos = (start+n-i)%240;
			sum += a_h[i-1]*s[pos];
		}
		pos = (start+n)%240;
		r[n] = s[pos] + sum;
	}

	//Compute the weighted speech signal s_w(n) for n = 0...39
	// Equation (A.2)
	for (n=0; n<40; n++)
	{
		float sum = 0;
		for (i=1; i<=10; i++)
		{			
//			if (n-i<0)
//				sum += a_p[i]*ew[10-i];
//			else
			{
				pos = (head+n-i)%320;
				sum += a_p[i]*s_w[pos];
			}
		}
		pos = (head+n)%320;
		s_w[pos] = r[n] - sum;
	}
}


/*********************************************************************
/*
/*	  Function to do Open loop pitch analysis [A.3.4]
/*
/*    Input: s_w[320], head
/*    Output: *t_op
/*
/*********************************************************************/
void open_loop_pitch_analysis(float s_w[320], int head, int *t_op)
{
	float R[144];
	int n;
	int k;
	int i;
	int pos, posk;
	int t[4];
	int posi[4];
	float Rt[4];
	float sum[4];
	
	head = head + 320;	

	for (i=1; i<4; i++)
	{
		Rt[i]  = 0;
		t[i]   = 0;
		sum[i] = 0;
		posi[i] = 0;
	}
	for (k=20; k<40; k++)
	{
		R[k] = 0;
		for (n=0; n<40; n++)
		{
			pos = (head+2*n)%320;
			posk = (head+2*n-k)%320;
			R[k] += s_w[pos]*s_w[posk];
		}
		if (k==20)
		{
			Rt[1] = R[k];
			t[1] = k;
		}
		else if (R[k]>Rt[1])
		{
			Rt[1] = R[k];
			t[1] = k;
		}
	}
	for (k=40; k<80; k++)
	{
		R[k] = 0;
		for (n=0; n<40; n++)
		{
			pos = (head+2*n)%320;
			posk = (head+2*n-k)%320;
			R[k] += s_w[pos]*s_w[posk];
		}
		if (k==40)
		{
			Rt[2] = R[k];
			t[2] = k;
		}
		else if (R[k]>Rt[2])
		{
			Rt[2] = R[k];
			t[2] = k;
		}
	}
	for (k=80; k<144; k++)
	{
		R[k] = 0;
		for (n=0; n<40; n++)
		{
			pos = (head+2*n)%320;
			posk = (head+2*n-k)%320;
			R[k] += s_w[pos]*s_w[posk];
		}
		if (k==80)
		{
			Rt[3] = R[k];
			t[3] = k;
		}
		else if (R[k]>Rt[3])
		{
			Rt[3] = R[k];
			t[3] = k;
		}
	}
	for (n=0; n<40; n++)
	{
		posi[1] = (head+2*n-t[1])%320;
		sum[1] += s_w[posi[1]]*s_w[posi[1]];
		posi[2] = (head+2*n-t[2])%320;
		sum[2] += s_w[posi[2]]*s_w[posi[2]];
		posi[3] = (head+2*n-t[3])%320;
		sum[3] += s_w[posi[3]]*s_w[posi[3]];
	}
	Rt[1] = Rt[1]/sqrt(sum[1]);
	Rt[2] = Rt[2]/sqrt(sum[2]);
	Rt[3] = Rt[3]/sqrt(sum[3]);

	if (Rt[1]>Rt[2] && Rt[1]>Rt[3])
		*t_op = t[1];
	else if (Rt[2]>Rt[1] && Rt[2]>Rt[3])
		*t_op = t[2];
	else 
		*t_op = t[3];
}

/*********************************************************************
/*
/*	  Function to compute the impulse response [A.3.5]
/*                compute the target signal [A.3.6]
/*
/*    Input : a_h[10]
/*		    : r[40]
/*			: ew[10]
/*    Output: h[40]
/*			: x[40]
/*
/****v*****************************************************************/
void response(float a_h[10], float r[40], float h[40], float x[40], float ew[10])
{
	int n, i;
	float gamma;
	float ir, ts;
	
	h[0] = 1;
	x[0] = r[0];
	for (n=1; n<40; n++)
	{
		ir = 0;
		ts = 0;
		gamma = .75;
		for (i=1; i<=10; i++)
		{
			if (n-i>=0)
			{
				ir += a_h[i-1]*gamma*h[n-i];
				ts += a_h[i-1]*gamma*x[n-i];
			}
			else
			{
				ts += a_h[i-1]*gamma*ew[10-i];
			}
			gamma *= .75;
		}
		h[n] = -ir;
		x[n] = r[n]-ts;
	}
}
